@@ -1,16 +1,5 @@
-// Copyright 2020 ConsenSys Software Inc.
-//
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//     http://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Copyright 2020-2024 Consensys Software Inc.
+// Licensed under the Apache License, Version 2.0. See the LICENSE file for details.
 
 // Package amd64 contains wrapper to amd64 instructions in Go assembly.
 // note that while this package is public, it is tailored for github.com/consensys/goff and github.com/consensys/gurvy
@@ -47,6 +36,62 @@ func (amd64 *Amd64) RET() {
 }
 
 // AVX 512 instructions
+// some of the documentation (in particular for AVX512 ops) is taken from https://github.com/mmcloughlin/avo
+
+// VPMULLD: Multiply Packed Signed Doubleword Integers and Store Low Result.
+func (amd64 *Amd64) VPMULLD(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMULLD", r1, r2, r3)
+}
+
+// VPMULLD_BCST: Multiply Packed Signed Doubleword Integers and Store Low Result (Broadcast).
+func (amd64 *Amd64) VPMULLD_BCST(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMULLD.BCST", r1, r2, r3)
+}
+
+// VPMULLQ: Multiply Packed Signed Quadword Integers and Store Low Result.
+func (amd64 *Amd64) VPMULLQ(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMULLQ", r1, r2, r3)
+}
+
+// VPMULLQ_BCST: Multiply Packed Signed Quadword Integers and Store Low Result (Broadcast).
+func (amd64 *Amd64) VPMULLQ_BCST(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMULLQ.BCST", r1, r2, r3)
+}
+
+// VMOVSHDUP: Move Packed Single-FP High and Duplicate
+func (amd64 *Amd64) VMOVSHDUP(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VMOVSHDUP", r1, r2)
+}
+
+// VPBROADCASTQ: Broadcast Quadword Integer
+func (amd64 *Amd64) VPBROADCASTQ(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPBROADCASTQ", r1, r2)
+}
+
+// VPBROADCASTD: Broadcast Doubleword Integer
+func (amd64 *Amd64) VPBROADCASTD(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPBROADCASTD", r1, r2)
+}
+
+// VPADDD: Add Packed Doubleword Integers
+func (amd64 *Amd64) VPADDD(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPADDD", r1, r2, r3)
+}
+
+// VPSUBD: Subtract Packed Doubleword Integers
+func (amd64 *Amd64) VPSUBD(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPSUBD", r1, r2, r3)
+}
+
+// VPMINUD: Minimum of Packed Unsigned Doubleword Integers
+func (amd64 *Amd64) VPMINUD(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMINUD", r1, r2, r3)
+}
+
+// VPMINUQ: Minimum of Packed Unsigned Quadword Integers
+func (amd64 *Amd64) VPMINUQ(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMINUQ", r1, r2, r3)
+}
 
 // VPSLLQ: Shift Packed Quadword Data Left Logical
 func (amd64 *Amd64) VPSLLQ(r1, r2, r3 interface{}, comment ...string) {
@@ -73,6 +118,95 @@ func (amd64 *Amd64) VPERMQ(r1, r2, r3 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VPERMQ", r1, r2, r3)
 }
 
+// VEXTRACTI32X8: Extract 256 Bits of Packed Doubleword Integer Values.
+//
+// Forms:
+//
+//	VEXTRACTI32X8 imm8 zmm k m256
+//	VEXTRACTI32X8 imm8 zmm k ymm
+//	VEXTRACTI32X8 imm8 zmm m256
+//	VEXTRACTI32X8 imm8 zmm ymm
+func (amd64 *Amd64) VEXTRACTI32X8(imm8, r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VEXTRACTI32X8", imm8, r1, r2)
+}
+
+// VPBLENDMQ: Blend Quadword Vectors Using an OpMask Control.
+//
+// Forms:
+//
+//	VPBLENDMQ m128 xmm k xmm
+//	VPBLENDMQ m128 xmm xmm
+//	VPBLENDMQ m256 ymm k ymm
+//	VPBLENDMQ m256 ymm ymm
+//	VPBLENDMQ xmm  xmm k xmm
+//	VPBLENDMQ xmm  xmm xmm
+//	VPBLENDMQ ymm  ymm k ymm
+//	VPBLENDMQ ymm  ymm ymm
+//	VPBLENDMQ m512 zmm k zmm
+//	VPBLENDMQ m512 zmm zmm
+//	VPBLENDMQ zmm  zmm k zmm
+//	VPBLENDMQ zmm  zmm zmm
+func (amd64 *Amd64) VPBLENDMQ(r1, r2, r3, k interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPBLENDMQ", r1, r2, k, r3)
+}
+
+// VEXTRACTI64X4: Extract 256 Bits of Packed Quadword Integer Values.
+//
+// Forms:
+//
+//	VEXTRACTI64X4 imm8 zmm k m256
+//	VEXTRACTI64X4 imm8 zmm k ymm
+//	VEXTRACTI64X4 imm8 zmm m256
+//	VEXTRACTI64X4 imm8 zmm ymm
+func (amd64 *Amd64) VEXTRACTI64X4(imm8, r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VEXTRACTI64X4", imm8, r1, r2)
+}
+
+// VPSHRDQ: Concatenate Quadwords and Shift Packed Data Right Logical.
+//
+// Forms:
+//
+//	VPSHRDQ imm8 m128 xmm k xmm
+//	VPSHRDQ imm8 m128 xmm xmm
+//	VPSHRDQ imm8 m256 ymm k ymm
+//	VPSHRDQ imm8 m256 ymm ymm
+//	VPSHRDQ imm8 xmm  xmm k xmm
+//	VPSHRDQ imm8 xmm  xmm xmm
+//	VPSHRDQ imm8 ymm  ymm k ymm
+//	VPSHRDQ imm8 ymm  ymm ymm
+//	VPSHRDQ imm8 m512 zmm k zmm
+//	VPSHRDQ imm8 m512 zmm zmm
+//	VPSHRDQ imm8 zmm  zmm k zmm
+//	VPSHRDQ imm8 zmm  zmm zmm
+func (amd64 *Amd64) VPSHRDQ(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPSHRDQ", imm8, r1, r2, r3)
+}
+
+// VPBLENDMD: Blend Doubleword Vectors Using an OpMask Control.
+//
+// Forms:
+//
+//	VPBLENDMD m128 xmm k xmm
+//	VPBLENDMD m128 xmm xmm
+//	VPBLENDMD m256 ymm k ymm
+//	VPBLENDMD m256 ymm ymm
+//	VPBLENDMD xmm  xmm k xmm
+//	VPBLENDMD xmm  xmm xmm
+//	VPBLENDMD ymm  ymm k ymm
+//	VPBLENDMD ymm  ymm ymm
+//	VPBLENDMD m512 zmm k zmm
+//	VPBLENDMD m512 zmm zmm
+//	VPBLENDMD zmm  zmm k zmm
+//	VPBLENDMD zmm  zmm zmm
+func (amd64 *Amd64) VPBLENDMD(r1, r2, r3, k interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPBLENDMD", r1, r2, k, r3)
+}
+
+// VEXTRACTI64X2
+func (amd64 *Amd64) VEXTRACTI64X2(imm8, r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VEXTRACTI64X2", imm8, r1, r2)
+}
+
 // VPERMD: Permute Doubleword Integers
 func (amd64 *Amd64) VPERMD(r1, r2, r3 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VPERMD", r1, r2, r3)
@@ -93,6 +227,21 @@ func (amd64 *Amd64) KMOVD(r1, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "KMOVD", r1, r2)
 }
 
+// KMOVQ Move 64-bit Mask
+func (amd64 *Amd64) KMOVQ(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "KMOVQ", r1, r2)
+}
+
+// VPINSRQ: Insert Quadword.
+//
+// Forms:
+//
+//	VPINSRQ imm8 m64 xmm xmm
+//	VPINSRQ imm8 r64 xmm xmm
+func (amd64 *Amd64) VPINSRQ(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPINSRQ", imm8, r1, r2, r3)
+}
+
 // KSHIFTLW Shift 16-bit Mask Left
 func (amd64 *Amd64) KSHIFTLW(r1, r2, r3 interface{}, comment ...string) {
 	amd64.writeOp(comment, "KSHIFTLW", r1, r2, r3)
@@ -128,6 +277,11 @@ func (amd64 *Amd64) VPMOVQ2M(r1, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VPMOVQ2M", r1, r2)
 }
 
+// VMOVDQA32: Move Aligned Doubleword Values
+func (amd64 *Amd64) VMOVDQA32(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VMOVDQA32", r1, r2)
+}
+
 // VMOVDQA64 Move Aligned Quadword Values
 func (amd64 *Amd64) VMOVDQA64(r1, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VMOVDQA64", r1, r2)
@@ -138,16 +292,130 @@ func (amd64 *Amd64) VMOVDQA64_Z(r1, k, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VMOVDQA64.Z", r1, k, r2)
 }
 
+// VPMOVQD: Down Convert Packed Quadword Values to Doubleword Values with Truncation.
+func (amd64 *Amd64) VPMOVQD(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPMOVQD", r1, r2)
+}
+
 // VPMOVZXDQ Move Packed Doubleword Integers to Quadword Integers
 func (amd64 *Amd64) VPMOVZXDQ(r1, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VPMOVZXDQ", r1, r2)
 }
 
+// VPSHUFD: Shuffle Packed Doublewords.
+//
+// Forms:
+//
+//	VPSHUFD imm8 m256 ymm
+//	VPSHUFD imm8 ymm  ymm
+//	VPSHUFD imm8 m128 xmm
+//	VPSHUFD imm8 xmm  xmm
+//	VPSHUFD imm8 m128 k xmm
+//	VPSHUFD imm8 m256 k ymm
+//	VPSHUFD imm8 xmm  k xmm
+//	VPSHUFD imm8 ymm  k ymm
+//	VPSHUFD imm8 m512 k zmm
+//	VPSHUFD imm8 m512 zmm
+//	VPSHUFD imm8 zmm  k zmm
+//	VPSHUFD imm8 zmm  zmm
+func (amd64 *Amd64) VPSHUFD(imm8, r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPSHUFD", imm8, r1, r2)
+}
+
+// VSHUFPD: Shuffle Packed Double-Precision Floating-Point Values.
+//
+// Forms:
+//
+//	VSHUFPD imm8 m128 xmm xmm
+//	VSHUFPD imm8 m256 ymm ymm
+//	VSHUFPD imm8 xmm  xmm xmm
+//	VSHUFPD imm8 ymm  ymm ymm
+//	VSHUFPD imm8 m128 xmm k xmm
+//	VSHUFPD imm8 m256 ymm k ymm
+//	VSHUFPD imm8 xmm  xmm k xmm
+//	VSHUFPD imm8 ymm  ymm k ymm
+//	VSHUFPD imm8 m512 zmm k zmm
+//	VSHUFPD imm8 m512 zmm zmm
+//	VSHUFPD imm8 zmm  zmm k zmm
+//	VSHUFPD imm8 zmm  zmm zmm
+func (amd64 *Amd64) VSHUFPD(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VSHUFPD", imm8, r1, r2, r3)
+}
+
+// VSHUFF32X4: Shuffle 128-Bit Packed Single-Precision Floating-Point Values.
+//
+// Forms:
+//
+//	VSHUFF32X4 imm8 m256 ymm k ymm
+//	VSHUFF32X4 imm8 m256 ymm ymm
+//	VSHUFF32X4 imm8 ymm  ymm k ymm
+//	VSHUFF32X4 imm8 ymm  ymm ymm
+//	VSHUFF32X4 imm8 m512 zmm k zmm
+//	VSHUFF32X4 imm8 m512 zmm zmm
+//	VSHUFF32X4 imm8 zmm  zmm k zmm
+//	VSHUFF32X4 imm8 zmm  zmm zmm
+func (amd64 *Amd64) VSHUFF32X4(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VSHUFF32X4", imm8, r1, r2, r3)
+}
+
+// VINSERTI64X4: Insert 256 Bits of Packed Quadword Integer Values.
+//
+// Forms:
+//
+//	VINSERTI64X4 imm8 m256 zmm k zmm
+//	VINSERTI64X4 imm8 m256 zmm zmm
+//	VINSERTI64X4 imm8 ymm  zmm k zmm
+//	VINSERTI64X4 imm8 ymm  zmm zmm
+func (amd64 *Amd64) VINSERTI64X4(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VINSERTI64X4", imm8, r1, r2, r3)
+}
+
+// VINSERTI64X2: Insert 128 Bits of Packed Quadword Integer Values.
+//
+// Forms:
+//
+//	VINSERTI64X2 imm8 m128 ymm k ymm
+//	VINSERTI64X2 imm8 m128 ymm ymm
+//	VINSERTI64X2 imm8 xmm  ymm k ymm
+//	VINSERTI64X2 imm8 xmm  ymm ymm
+//	VINSERTI64X2 imm8 m128 zmm k zmm
+//	VINSERTI64X2 imm8 m128 zmm zmm
+//	VINSERTI64X2 imm8 xmm  zmm k zmm
+//	VINSERTI64X2 imm8 xmm  zmm zmm
+func (amd64 *Amd64) VINSERTI64X2(imm8, r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VINSERTI64X2", imm8, r1, r2, r3)
+}
+
+// VMOVDQU32: Move Unaligned Doubleword Values
+func (amd64 *Amd64) VMOVDQU32(r1, r2 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VMOVDQU32", r1, r2)
+}
+
 // VMOVDQU64 Move Unaligned Quadword Values
 func (amd64 *Amd64) VMOVDQU64(r1, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VMOVDQU64", r1, r2)
 }
 
+// VPERMI2Q: Full Permute of Quadwords From Two Tables Overwriting the Index.
+//
+// Forms:
+//
+//	VPERMI2Q m128 xmm k xmm
+//	VPERMI2Q m128 xmm xmm
+//	VPERMI2Q m256 ymm k ymm
+//	VPERMI2Q m256 ymm ymm
+//	VPERMI2Q xmm  xmm k xmm
+//	VPERMI2Q xmm  xmm xmm
+//	VPERMI2Q ymm  ymm k ymm
+//	VPERMI2Q ymm  ymm ymm
+//	VPERMI2Q m512 zmm k zmm
+//	VPERMI2Q m512 zmm zmm
+//	VPERMI2Q zmm  zmm k zmm
+//	VPERMI2Q zmm  zmm zmm
+func (amd64 *Amd64) VPERMI2Q(r1, r2, r3 interface{}, comment ...string) {
+	amd64.writeOp(comment, "VPERMI2Q", r1, r2, r3)
+}
+
 // VMOVDQU64 Move Unaligned Quadword Values
 func (amd64 *Amd64) VMOVDQU64k(r1, k, r2 interface{}, comment ...string) {
 	amd64.writeOp(comment, "VMOVDQU64", r1, k, r2)
@@ -379,6 +647,11 @@ func (amd64 *Amd64) JCC(label Label, comment ...string) {
 	amd64.writeOp(comment, "JCC", string(label))
 }
 
+// JGE
+func (amd64 *Amd64) JGE(label Label, comment ...string) {
+	amd64.writeOp(comment, "JGE", string(label))
+}
+
 func (amd64 *Amd64) JMP(label Label, comment ...string) {
 	amd64.writeOp(comment, "JMP", string(label))
 }
