@@ -1,14 +1,11 @@
 package fptower
 
 import (
-	"github.com/consensys/gnark-crypto/ecc/bls12-381/fp"
+	"github.com/consensys/gnark-crypto/ecc/bls24-317/fp"
 	"github.com/leanovate/gopter"
 )
 
-// ------------------------------------------------------------
-// Tower generators
-
-// GenFp generates an Fp element
+// Fp generates an Fp element
 func GenFp() gopter.Gen {
 	return func(genParams *gopter.GenParameters) *gopter.GenResult {
 		var elmt fp.Element
@@ -21,7 +18,7 @@ func GenFp() gopter.Gen {
 	}
 }
 
-// GenE2 generates an E2 elmt
+// E2 generates an E2 elmt
 func GenE2() gopter.Gen {
 	return gopter.CombineGens(
 		GenFp(),
@@ -31,23 +28,33 @@ func GenE2() gopter.Gen {
 	})
 }
 
-// GenE6 generates an E6 elmt
-func GenE6() gopter.Gen {
+// E4 generates an E4 elmt
+func GenE4() gopter.Gen {
 	return gopter.CombineGens(
 		GenE2(),
 		GenE2(),
-		GenE2(),
-	).Map(func(values []interface{}) *E6 {
-		return &E6{B0: *values[0].(*E2), B1: *values[1].(*E2), B2: *values[2].(*E2)}
+	).Map(func(values []interface{}) *E4 {
+		return &E4{B0: *values[0].(*E2), B1: *values[1].(*E2)}
 	})
 }
 
-// GenE12 generates an E6 elmt
+// E12 generates an E12 elmt
 func GenE12() gopter.Gen {
 	return gopter.CombineGens(
-		GenE6(),
-		GenE6(),
+		GenE4(),
+		GenE4(),
+		GenE4(),
 	).Map(func(values []interface{}) *E12 {
-		return &E12{C0: *values[0].(*E6), C1: *values[1].(*E6)}
+		return &E12{C0: *values[0].(*E4), C1: *values[1].(*E4), C2: *values[2].(*E4)}
+	})
+}
+
+// E24 generates an E24 elmt
+func GenE24() gopter.Gen {
+	return gopter.CombineGens(
+		GenE12(),
+		GenE12(),
+	).Map(func(values []interface{}) *E24 {
+		return &E24{D0: *values[0].(*E12), D1: *values[1].(*E12)}
 	})
 }
