@@ -18,7 +18,7 @@ import (
 	"github.com/bits-and-blooms/bitset"
 )
 
-// {{.ElementName}} represents a field element stored on {{.NbWords}} words (uint64)
+// {{.ElementName}} represents a field element stored on {{.NbWords}} words ({{$.Word.TypeLower}})
 //
 // {{.ElementName}} are assumed to be in Montgomery form in all methods.
 //
@@ -30,10 +30,10 @@ import (
 // Warning
 //
 // This code has not been audited and is provided as-is. In particular, there is no security guarantees such as constant time implementation or side-channel attack resistance.
-type {{.ElementName}} [{{.NbWords}}]uint64
+type {{.ElementName}} [{{.NbWords}}]{{$.Word.TypeLower}}
 
 const (
-	Limbs = {{.NbWords}} 	// number of 64 bits words needed to represent a {{.ElementName}}
+	Limbs = {{.NbWords}} 	// number of {{$.Word.BitSize}} bits words needed to represent a {{.ElementName}}
 	Bits = {{.NbBits}} 		// number of bits needed to represent a {{.ElementName}}
 	Bytes = {{.NbBytes}} 	// number of bytes needed to represent a {{.ElementName}}
 )
@@ -42,14 +42,14 @@ const (
 // Field modulus q
 const (
 {{- range $i := $.NbWordsIndexesFull}}
-	q{{$i}} uint64 = {{index $.Q $i}}
+	q{{$i}} = {{index $.Q $i}}
 	{{- if eq $.NbWords 1}}
-		q uint64 = q0
+		q  = q0
 	{{- end}}
 {{- end}}
 )
 
-var q{{.ElementName}} = {{.ElementName}}{
+var qElement = {{.ElementName}}{
 	{{- range $i := $.NbWordsIndexesFull}}
 	q{{$i}},{{end}}
 }
@@ -66,7 +66,13 @@ func Modulus() *big.Int {
 
 // q + r'.r = 1, i.e., qInvNeg = - q⁻¹ mod r
 // used for Montgomery reduction
-const qInvNeg uint64 = {{index .QInverse 0}}
+const qInvNeg = {{index .QInverse 0}}
+
+{{- if eq .NbWords 4}}
+// mu = 2^288 / q needed for partial Barrett reduction
+const mu uint64 = {{.Mu}}
+{{- end}}
+
 
 func init() {
 	_modulus.SetString("{{.ModulusHex}}", 16)
@@ -78,16 +84,27 @@ func init() {
 // 		var v {{.ElementName}}
 // 		v.SetUint64(...)
 func New{{.ElementName}}(v uint64) {{.ElementName}} {
-	z := {{.ElementName}}{v}
-	z.Mul(&z, &rSquare)
-	return z
+	{{- if .F31}}
+		z := {{.ElementName}}{ uint32(v % uint64(q0)) }
+		z.toMont()
+		return z
+	{{- else }}
+		z := {{.ElementName}}{ v }
+		z.Mul(&z, &rSquare)
+		return z
+	{{- end}}
 }
 
 // SetUint64 sets z to v and returns z
 func (z *{{.ElementName}}) SetUint64(v uint64) *{{.ElementName}} {
 	//  sets z LSB to v (non-Montgomery form) and convert z to Montgomery form
-	*z = {{.ElementName}}{v}
-	return z.Mul(z, &rSquare) // z.toMont()
+	{{- if .F31}}
+		*z = {{.ElementName}}{ uint32(v % uint64(q0)) }
+		return z.toMont()
+	{{- else }}
+		*z = {{.ElementName}}{ v }
+		return z.Mul(z, &rSquare) // z.toMont()
+	{{- end}}
 }
 
 // SetInt64 sets z to v and returns z
@@ -204,7 +221,7 @@ func (z *{{.ElementName}}) Equal(x *{{.ElementName}}) bool {
 }
 
 // NotEqual returns 0 if and only if z == x; constant-time
-func (z *{{.ElementName}}) NotEqual(x *{{.ElementName}}) uint64 {
+func (z *{{.ElementName}}) NotEqual(x *{{.ElementName}}) {{.Word.TypeLower}} {
 return {{- range $i :=  reverse .NbWordsIndexesNoZero}}(z[{{$i}}] ^ x[{{$i}}]) | {{end}}(z[0] ^ x[0])
 }
 
@@ -235,7 +252,11 @@ func (z *{{.ElementName}}) IsUint64() bool {
 
 // Uint64 returns the uint64 representation of x. If x cannot be represented in a uint64, the result is undefined.
 func (z *{{.ElementName}}) Uint64() uint64 {
-	return z.Bits()[0]
+	{{- if eq .Word.BitSize 32}}
+		return uint64(z.Bits()[0])
+	{{- else}}
+		return z.Bits()[0]
+	{{- end}}
 }
 
 // FitsOnOneWord reports whether z words (except the least significant word) are 0
@@ -277,10 +298,10 @@ func (z *{{.ElementName}}) LexicographicallyLargest() bool {
 
 	_z := z.Bits()
 
-	var b uint64
-	_, b = bits.Sub64(_z[0], {{index .QMinusOneHalvedP 0}}, 0)
+	var b {{$.Word.TypeLower}}
+	_, b = bits.{{$.Word.Sub}}(_z[0], {{index .QMinusOneHalvedP 0}}, 0)
 	{{- range $i := .NbWordsIndexesNoZero}}
-		_, b = bits.Sub64(_z[{{$i}}], {{index $.QMinusOneHalvedP $i}}, b)
+		_, b = bits.{{$.Word.Sub}}(_z[{{$i}}], {{index $.QMinusOneHalvedP $i}}, b)
 	{{- end}}
 
 	return b == 0
@@ -323,7 +344,7 @@ func (z *{{.ElementName}}) SetRandom() (*{{.ElementName}}, error) {
 
 		{{- range $i :=  .NbWordsIndexesFull}}
 			{{- $k := add $i 1}}
-			z[{{$i}}] = binary.LittleEndian.Uint64(bytes[{{mul $i 8}}:{{mul $k 8}}])
+			z[{{$i}}] = binary.LittleEndian.{{$.Word.TypeUpper}}(bytes[{{mul $i $.Word.ByteSize}}:{{mul $k $.Word.ByteSize}}])
 		{{- end}}
 
 		if !z.smallerThanModulus() {
@@ -377,7 +398,7 @@ func (z *{{.ElementName}}) Halve()  {
 	{{- range $i := $.all.NbWordsIndexesFull }}
 		{{- $carryIn := ne $i 0}}
 		{{- $carryOut := or (ne $i $.all.NbWordsLastIndex) (and (eq $i $.all.NbWordsLastIndex) (not $.all.NoCarry))}}
-		{{$.V1}}[{{$i}}], {{- if $carryOut}}carry{{- else}}_{{- end}} = bits.Add64({{$.V1}}[{{$i}}], q{{$i}}, {{- if $carryIn}}carry{{- else}}0{{- end}})
+		{{$.V1}}[{{$i}}], {{- if $carryOut}}carry{{- else}}_{{- end}} = bits.{{$.all.Word.Add}}({{$.V1}}[{{$i}}], q{{$i}}, {{- if $carryIn}}carry{{- else}}0{{- end}})
 	{{- end}}
 {{ end }}
 
@@ -392,20 +413,31 @@ func (z *{{.ElementName}}) fromMont() *{{.ElementName}} {
 
 // Add z = x + y (mod q)
 func (z *{{.ElementName}}) Add( x, y *{{.ElementName}}) *{{.ElementName}} {
-	{{ $hasCarry := or (not $.NoCarry) (gt $.NbWords 1)}}
-	{{- if $hasCarry}}
-		var carry uint64
-	{{- end}}
-	{{- range $i := iterate 0 $.NbWords}}
-		{{- $hasCarry := or (not $.NoCarry) (lt $i $.NbWordsLastIndex)}}
-		z[{{$i}}], {{- if $hasCarry}}carry{{- else}}_{{- end}} = bits.Add64(x[{{$i}}], y[{{$i}}], {{- if eq $i 0}}0{{- else}}carry{{- end}})
-	{{- end}}
-
-	{{- if eq $.NbWords 1}}
-		if {{- if not .NoCarry}} carry != 0 ||{{- end }} z[0] >= q {
-			z[0] -= q
-		}
+	{{- if eq .NbWords 1}}
+		{{ $hasCarry := (not $.NoCarry)}}
+		{{- if $hasCarry}}
+			var carry uint64
+			z[0], carry = bits.Add64(x[0], y[0], 0)
+			if carry != 0 || z[0] >= q {
+				z[0] -= q
+			}
+			return z
+		{{- else}}
+			t := x[0] + y[0]
+			if t >= q {
+				t -= q
+			}
+			z[0] = t
+			return z
+		{{- end}}
 	{{- else}}
+	
+		var carry uint64
+		{{- range $i := iterate 0 $.NbWords}}
+			{{- $hasCarry := or (not $.NoCarry) (lt $i $.NbWordsLastIndex)}}
+			z[{{$i}}], {{- if $hasCarry}}carry{{- else}}_{{- end}} = bits.Add64(x[{{$i}}], y[{{$i}}], {{- if eq $i 0}}0{{- else}}carry{{- end}})
+		{{- end}}
+
 		{{- if not .NoCarry}}
 			// if we overflowed the last addition, z >= q
 			// if z >= q, z = z - q
@@ -419,25 +451,36 @@ func (z *{{.ElementName}}) Add( x, y *{{.ElementName}}) *{{.ElementName}} {
 				return z
 			}
 		{{- end}}
-
 		{{ template "reduce" .}}
+		 return z
 	{{- end}}
-	return z
 }
 
+
+
 // Double z = x + x (mod q), aka Lsh 1
 func (z *{{.ElementName}}) Double( x *{{.ElementName}}) *{{.ElementName}} {
 	{{- if eq .NbWords 1}}
-	if x[0] & (1 << 63) == (1 << 63) {
-		// if highest bit is set, then we have a carry to x + x, we shift and subtract q
-		z[0] = (x[0] << 1) - q
-	} else {
-		// highest bit is not set, but x + x can still be >= q
-		z[0] = (x[0] << 1)
-		if z[0] >= q {
-			z[0] -= q
-		}
-	}
+		{{- if .F31}}
+			t := x[0] << 1
+			if t >= q {
+					t -= q
+			}
+			z[0] = t
+			return z
+		{{- else}}
+			if x[0]&(1<<63) == (1 << 63) {
+				// if highest bit is set, then we have a carry to x + x, we shift and subtract q
+				z[0] = (x[0] << 1) - q
+			} else {
+				// highest bit is not set, but x + x can still be >= q
+				z[0] = (x[0] << 1)
+				if z[0] >= q {
+						z[0] -= q
+				}
+			}
+			return z
+		{{- end}}
 	{{- else}}
 	{{ $hasCarry := or (not $.NoCarry) (gt $.NbWords 1)}}
 	{{- if $hasCarry}}
@@ -462,36 +505,47 @@ func (z *{{.ElementName}}) Double( x *{{.ElementName}}) *{{.ElementName}} {
 	{{- end}}
 
 	{{ template "reduce" .}}
-	{{- end}}
 	return z
+	{{- end}}
 }
 
 
+
 // Sub z = x - y (mod q)
 func (z *{{.ElementName}}) Sub( x, y *{{.ElementName}}) *{{.ElementName}} {
-	var b uint64
-	z[0], b = bits.Sub64(x[0], y[0], 0)
-	{{- range $i := .NbWordsIndexesNoZero}}
-		z[{{$i}}], b = bits.Sub64(x[{{$i}}], y[{{$i}}], b)
-	{{- end}}
-	if b != 0 {
-		{{- if eq .NbWords 1}}
-			z[0] += q
-		{{- else}}
-			var c uint64
-			z[0], c = bits.Add64(z[0], q0, 0)
-			{{- range $i := .NbWordsIndexesNoZero}}
-				{{- if eq $i $.NbWordsLastIndex}}
-					z[{{$i}}], _ = bits.Add64(z[{{$i}}], q{{$i}}, c)
-				{{- else}}
-					z[{{$i}}], c = bits.Add64(z[{{$i}}], q{{$i}}, c)
+	{{- if $.F31}}
+		t, b := bits.Sub32(x[0], y[0], 0)
+		if b != 0 {
+			t += q
+		}
+		z[0] = t
+		return z
+	{{- else}}
+		var b uint64
+		z[0], b = bits.Sub64(x[0], y[0], 0)
+		{{- range $i := .NbWordsIndexesNoZero}}
+			z[{{$i}}], b = bits.Sub64(x[{{$i}}], y[{{$i}}], b)
+		{{- end}}
+		if b != 0 {
+			{{- if eq .NbWords 1}}
+				z[0] += q
+			{{- else}}
+				var c uint64
+				z[0], c = bits.Add64(z[0], q0, 0)
+				{{- range $i := .NbWordsIndexesNoZero}}
+					{{- if eq $i $.NbWordsLastIndex}}
+						z[{{$i}}], _ = bits.Add64(z[{{$i}}], q{{$i}}, c)
+					{{- else}}
+						z[{{$i}}], c = bits.Add64(z[{{$i}}], q{{$i}}, c)
+					{{- end}}
 				{{- end}}
 			{{- end}}
-		{{- end}}
-	}
-	return z
+		}
+		return z
+	{{- end}}
 }
 
+
 // Neg z = q - x
 func (z *{{.ElementName}}) Neg( x *{{.ElementName}}) *{{.ElementName}} {
 	if x.IsZero() {
@@ -514,16 +568,18 @@ func (z *{{.ElementName}}) Neg( x *{{.ElementName}}) *{{.ElementName}} {
 	return z
 }
 
+
 // Select is a constant-time conditional move.
 // If c=0, z = x0. Else z = x1
 func (z *{{.ElementName}}) Select(c int, x0 *{{.ElementName}}, x1 *{{.ElementName}}) *{{.ElementName}} {
-	cC := uint64( (int64(c) | -int64(c)) >> 63 )	// "canonicized" into: 0 if c=0, -1 otherwise
+	cC := {{$.Word.TypeLower}}( (int64(c) | -int64(c)) >> 63 )	// "canonicized" into: 0 if c=0, -1 otherwise
 	{{- range $i := .NbWordsIndexesFull }}
 	z[{{$i}}] = x0[{{$i}}] ^ cC & (x0[{{$i}}] ^ x1[{{$i}}])
 	{{- end}}
 	return z
 }
 
+{{- if ne .NbWords 1}}
 // _mulGeneric is unoptimized textbook CIOS
 // it is a fallback solution on x86 when ADX instruction set is not available
 // and is used for testing purposes.
@@ -532,25 +588,30 @@ func _mulGeneric(z,x,y *{{.ElementName}}) {
 	{{ template "mul_cios" dict "all" . "V1" "x" "V2" "y"}}
 	{{ template "reduce"  . }}
 }
+{{- end}}
 
 
 func _fromMontGeneric(z *{{.ElementName}}) {
-	// the following lines implement z = z * 1
-	// with a modified CIOS montgomery multiplication
-	// see Mul for algorithm documentation
-	{{- range $j := .NbWordsIndexesFull}}
-	{
-		// m = z[0]n'[0] mod W
-		m := z[0] * qInvNeg
-		C := madd0(m, q0, z[0])
-		{{- range $i := $.NbWordsIndexesNoZero}}
-			C, z[{{sub $i 1}}] = madd2(m, q{{$i}}, z[{{$i}}], C)
+	{{- if .F31}}
+		z[0] = montReduce(uint64(z[0]))
+	{{- else}}
+		// the following lines implement z = z * 1
+		// with a modified CIOS montgomery multiplication
+		// see Mul for algorithm documentation
+		{{- range $j := .NbWordsIndexesFull}}
+		{
+			// m = z[0]n'[0] mod W
+			m := z[0] * qInvNeg
+			C := madd0(m, q0, z[0])
+			{{- range $i := $.NbWordsIndexesNoZero}}
+				C, z[{{sub $i 1}}] = madd2(m, q{{$i}}, z[{{$i}}], C)
+			{{- end}}
+			z[{{sub $.NbWords 1}}] = C
+		}
 		{{- end}}
-		z[{{sub $.NbWords 1}}] = C
-	}
-	{{- end}}
 
-	{{ template "reduce" .}}
+		{{ template "reduce" .}}
+	{{- end}}
 }
 
 func _reduceGeneric(z *{{.ElementName}})  {
@@ -601,10 +662,10 @@ func _butterflyGeneric(a, b *{{.ElementName}}) {
 func (z *{{.ElementName}}) BitLen() int {
 	{{- range $i := reverse .NbWordsIndexesNoZero}}
 	if z[{{$i}}] != 0 {
-		return {{mul $i 64}} + bits.Len64(z[{{$i}}])
+		return {{mul $i 64}} + bits.{{$.Word.Len}}(z[{{$i}}])
 	}
 	{{- end}}
-	return bits.Len64(z[0])
+	return bits.{{$.Word.Len}}(z[0])
 }
 
 // Hash msg to count prime field elements.
