@@ -207,6 +207,7 @@ func Benchmark{{toTitle .ElementName}}Mul(b *testing.B) {
 	}
 }
 
+
 func Benchmark{{toTitle .ElementName}}Cmp(b *testing.B) {
 	x := {{.ElementName}}{
 		{{- range $i := .RSquare}}
@@ -331,14 +332,14 @@ func init() {
 	staticTestValues = append(staticTestValues, rSquare) 			// r²
 	var e, one {{.ElementName}}
 	one.SetOne()
-	e.Sub(&q{{.ElementName}}, &one)
+	e.Sub(&qElement, &one)
 	staticTestValues = append(staticTestValues, e) 	// q - 1
 	e.Double(&one)
 	staticTestValues = append(staticTestValues, e) 	// 2 
 
 
 	{
-		a := q{{.ElementName}}
+		a := qElement
 		a[0]--
 		staticTestValues = append(staticTestValues, a)
 	}
@@ -354,14 +355,14 @@ func init() {
 	{{- end}}
 
 	{
-		a := q{{.ElementName}}
+		a := qElement
 		a[{{.NbWordsLastIndex}}]--
 		staticTestValues = append(staticTestValues, a)
 	}
 
 	{{- if ne .NbWords 1}}
 	{
-		a := q{{.ElementName}}
+		a := qElement
 		a[{{.NbWordsLastIndex}}]--
 		a[0]++
 		staticTestValues = append(staticTestValues, a)
@@ -369,7 +370,7 @@ func init() {
 	{{- end}}
 
 	{
-		a := q{{.ElementName}}
+		a := qElement
 		a[{{.NbWordsLastIndex}}] = 0
 		staticTestValues = append(staticTestValues, a)
 	}
@@ -653,8 +654,6 @@ func Test{{toTitle .ElementName}}BitLen(t *testing.T) {
 	))
 
 	properties.TestingRun(t, gopter.ConsoleReporter(false))
-
-	
 }
 
 
@@ -731,7 +730,11 @@ func Test{{toTitle .ElementName}}LexicographicallyLargest(t *testing.T) {
 
 {{template "testBinaryOp" dict "all" . "Op" "Add"}}
 {{template "testBinaryOp" dict "all" . "Op" "Sub"}}
+{{- if ne .NbWords 1}}
 {{template "testBinaryOp" dict "all" . "Op" "Mul" "GenericOp" "_mulGeneric"}}
+{{- else}}
+{{template "testBinaryOp" dict "all" . "Op" "Mul"}}
+{{- end}}
 {{template "testBinaryOp" dict "all" . "Op" "Div"}}
 {{template "testBinaryOp" dict "all" . "Op" "Exp"}}
 
@@ -1570,6 +1573,10 @@ type testPair{{.ElementName}} struct {
 	bigint       big.Int
 }
 
+{{- $gen64 := "genParams.NextUint64()"}}
+{{- if eq .Word.BitSize 32}}
+{{- $gen64 = "uint32(genParams.NextUint64())"}}
+{{- end}}
 
 func gen() gopter.Gen {
 	return func(genParams *gopter.GenParameters) *gopter.GenResult {
@@ -1577,20 +1584,20 @@ func gen() gopter.Gen {
 
 		g.element = {{.ElementName}}{
 			{{- range $i := .NbWordsIndexesFull}}
-			genParams.NextUint64(),{{end}}
+			{{$gen64}},{{end}}
 		}
-		if q{{.ElementName}}[{{.NbWordsLastIndex}}] != ^uint64(0) {
-			g.element[{{.NbWordsLastIndex}}] %= (q{{.ElementName}}[{{.NbWordsLastIndex}}] +1 )
+		if qElement[{{.NbWordsLastIndex}}] != ^{{$.Word.TypeLower}}(0) {
+			g.element[{{.NbWordsLastIndex}}] %= (qElement[{{.NbWordsLastIndex}}] +1 )
 		}
 		
 
 		for !g.element.smallerThanModulus() {
 			g.element = {{.ElementName}}{
 				{{- range $i := .NbWordsIndexesFull}}
-				genParams.NextUint64(),{{end}}
+				{{$gen64}},{{end}}
 			}
-			if q{{.ElementName}}[{{.NbWordsLastIndex}}] != ^uint64(0) {
-				g.element[{{.NbWordsLastIndex}}] %= (q{{.ElementName}}[{{.NbWordsLastIndex}}] +1 )
+			if qElement[{{.NbWordsLastIndex}}] != ^{{$.Word.TypeLower}}(0) {
+				g.element[{{.NbWordsLastIndex}}] %= (qElement[{{.NbWordsLastIndex}}] +1 )
 			}
 		}
 
@@ -1600,42 +1607,42 @@ func gen() gopter.Gen {
 	}
 }
 
+func genRandomFq(genParams *gopter.GenParameters) {{.ElementName}} {
+	var g {{.ElementName}}
+
+	g = {{.ElementName}}{
+		{{- range $i := .NbWordsIndexesFull}}
+		{{$gen64}},{{end}}
+	}
+
+	if qElement[{{.NbWordsLastIndex}}] != ^{{$.Word.TypeLower}}(0) {
+		g[{{.NbWordsLastIndex}}] %= (qElement[{{.NbWordsLastIndex}}] +1 )
+	}
+
+	for !g.smallerThanModulus() {
+		g = {{.ElementName}}{
+			{{- range $i := .NbWordsIndexesFull}}
+			{{$gen64}},{{end}}
+		}
+		if qElement[{{.NbWordsLastIndex}}] != ^{{$.Word.TypeLower}}(0) {
+			g[{{.NbWordsLastIndex}}] %= (qElement[{{.NbWordsLastIndex}}] +1 )
+		}
+	}
+
+	return g 
+}
+
 
 func genFull() gopter.Gen {
 	return func(genParams *gopter.GenParameters) *gopter.GenResult {
+		a := genRandomFq(genParams)
 
-		genRandomFq := func() {{.ElementName}} {
-			var g {{.ElementName}}
-
-			g = {{.ElementName}}{
-				{{- range $i := .NbWordsIndexesFull}}
-				genParams.NextUint64(),{{end}}
-			}
-
-			if q{{.ElementName}}[{{.NbWordsLastIndex}}] != ^uint64(0) {
-				g[{{.NbWordsLastIndex}}] %= (q{{.ElementName}}[{{.NbWordsLastIndex}}] +1 )
-			}
-
-			for !g.smallerThanModulus() {
-				g = {{.ElementName}}{
-					{{- range $i := .NbWordsIndexesFull}}
-					genParams.NextUint64(),{{end}}
-				}
-				if q{{.ElementName}}[{{.NbWordsLastIndex}}] != ^uint64(0) {
-					g[{{.NbWordsLastIndex}}] %= (q{{.ElementName}}[{{.NbWordsLastIndex}}] +1 )
-				}
-			}
-
-			return g 
-		}
-		a := genRandomFq()
-
-		var carry uint64
+		var carry {{$.Word.TypeLower}}
 		{{- range $i := .NbWordsIndexesFull}}
 			{{- if eq $i $.NbWordsLastIndex}}
-			a[{{$i}}], _ = bits.Add64(a[{{$i}}], q{{$.ElementName}}[{{$i}}], carry)
+			a[{{$i}}], _ = bits.{{$.Word.Add}}(a[{{$i}}], qElement[{{$i}}], carry)
 			{{- else}}
-			a[{{$i}}], carry = bits.Add64(a[{{$i}}], q{{$.ElementName}}[{{$i}}], carry)
+			a[{{$i}}], carry = bits.{{$.Word.Add}}(a[{{$i}}], qElement[{{$i}}], carry)
 			{{- end}}
 		{{- end}}
 		
@@ -1643,6 +1650,15 @@ func genFull() gopter.Gen {
 		return genResult
 	}
 }
+
+func gen{{.ElementName}}() gopter.Gen {
+	return func(genParams *gopter.GenParameters) *gopter.GenResult {
+		a := genRandomFq(genParams)
+		genResult := gopter.NewGenResult(a, gopter.NoShrinker)
+		return genResult
+	}
+}
+
 {{if $.UsingP20Inverse}}
 func (z *{{.ElementName}}) matchVeryBigInt(aHi uint64, aInt *big.Int) error {
 	var modulus big.Int
